/*	Subtitler - Sub Station Alpha v2.x/4.x subtitling plugin for VirtualDub
	Copyright (C) 2000-2002 Avery Lee, All Rights Reserved.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	-----------------

	AntialiasMask.cpp: Collapses 8x monochrome masks into antialiased maps and
					   applies them with color
*/

#include "VBitmap.h"
#include "AntialiasMask.h"

extern unsigned char bitcnt[256];

AntialiasMask::AntialiasMask(int w, int h) {
	pBits = new unsigned char[w*h*2];

	memset(pBits, 0, w*h*2);

	this->w = w;
	this->h = h;
}

AntialiasMask::~AntialiasMask() {
	delete[] pBits;
}

void AntialiasMask::CreateAlpha(unsigned char *pBits, int pitch, int xborder, int yborder, bool fBackground, unsigned char *pGammaRamp) {
	int x, y;
	unsigned char *dst = this->pBits + (this->w*yborder + xborder)*2;

	if (fBackground)
		++dst;

	for(y=0; y<h-yborder*2; y++) {
		for(x=0; x<w-xborder*2; x++) {
			int alpha;

			alpha  = bitcnt[pBits[x+0*pitch]];
			alpha += bitcnt[pBits[x+1*pitch]];
			alpha += bitcnt[pBits[x+2*pitch]];
			alpha += bitcnt[pBits[x+3*pitch]];
			alpha += bitcnt[pBits[x+4*pitch]];
			alpha += bitcnt[pBits[x+5*pitch]];
			alpha += bitcnt[pBits[x+6*pitch]];
			alpha += bitcnt[pBits[x+7*pitch]];

			*dst = pGammaRamp[alpha];
			dst += 2;
		}

		pBits += 8*pitch;
		dst += 4*xborder;
	}
}

void AntialiasMask::Apply(const VBitmap *vbm, int x, int y, int w, int h, long back, long *switchpts) {
	int rbback = ((back<<16) & 0xff0000) | ((back>>16) & 0x0000ff);
	int gback = back & 0xff00;
	int rbfore, gfore;

	Pixel32 *dst;
	unsigned char *src;
	long srcmod, srcpitch, dstmod;
	int xo=0, yo=0;

	// clip!

	srcpitch = this->w;
	dstmod = (vbm->pitch>>2);

	if (x < 0) {
		xo=-x;
		w += x;
		x = 0;
	}
	if (y < 0) {
		h += y;
		y = 0;
	}

	if (x+w >= vbm->w)
		w = vbm->w - x;

	if (y+h >= vbm->h) {
		yo = y+h-vbm->h;
		h = vbm->h - y;
	}

	src = pBits + xo + (this->h - h - yo)*this->w;
	dst = vbm->Address32(x, y+h-1);

	srcmod = this->w - w;
	dstmod -= w;

	if (h>0 && w>0) do {
		int wt = w;
		long *switchptr = switchpts;
		Pixel32 *dst_switchpt = dst;

		do {
			int alpha0, alpha1, alpha2;

			while(dst >= dst_switchpt) {
				rbfore = ((*switchptr<<16)&0xff0000) | ((*switchptr>>16) & 0x0000ff);
				gfore = *switchptr++ & 0xff00;

				dst_switchpt += *switchptr++;
			}

			alpha1 = src[0];
			alpha2 = src[1];
			src+=2;

			alpha0 = 64 - alpha2;
			alpha2 -= alpha1;

			*dst = ((((*dst & 0xff00ff) * alpha0 + rbfore * alpha1 + rbback * alpha2 + 0x200020) >> 6)&0xff00ff)
				 + ((((*dst & 0x00ff00) * alpha0 +  gfore * alpha1 +  gback * alpha2 + 0x002000) >> 6)&0x00ff00);

			++dst;
		} while(--wt);

		dst += dstmod;
		src += srcmod*2;
	} while(--h);
}

void AntialiasMask::ApplyShadow(const VBitmap *vbm, int x, int y, int w, int h, long back) {
	int rbback = ((back<<16) & 0xff0000) | ((back>>16) & 0x0000ff);
	int gback = back & 0xff00;

	Pixel32 *dst;
	unsigned char *src;
	long srcmod, srcpitch, dstmod;
	int xo=0, yo=0;

	// clip!

	srcpitch = this->w;
	dstmod = (vbm->pitch>>2);

	if (x < 0) {
		xo=-x;
		w += x;
		x = 0;
	}
	if (y < 0) {
		h += y;
		y = 0;
	}

	if (x+w >= vbm->w)
		w = vbm->w - x;

	if (y+h >= vbm->h) {
		yo = y+h-vbm->h;
		h = vbm->h - y;
	}

	src = pBits + xo + (this->h - h + yo)*this->w;
	dst = vbm->Address32(x, y+h-1);

	srcmod = this->w - w;
	dstmod -= w;

	if (h>0 && w>0) do {
		int wt = w;

		do {
			int alpha0, alpha2;

			alpha2 = src[1];
			src+=2;

			alpha0 = 128 - alpha2;

			*dst = ((((*dst & 0xff00ff) * alpha0 + rbback * alpha2 + 0x400040) >> 7)&0xff00ff)
				 + ((((*dst & 0x00ff00) * alpha0 +  gback * alpha2 + 0x004000) >> 7)&0x00ff00);
			++dst;
		} while(--wt);

		dst += dstmod;
		src += srcmod*2;
	} while(--h);
}

