/*
	ChromaShift plugin for Avisynth -- lets users shift the chroma information
	around indenpendatly of the lunimance information

	Copyright (C) 2002 Simon Walters
  
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.
	 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
		
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The author can be contacted at:
	Simon Walters
	siwalters@hotmail.com
			 
	V1.0 10th Sep 2002 - 1st release
	V1.1 Allow variable amount of shift and independant U V shifts
	V1.2 Allow negative values and a global vertical shift value L
*/
#include "avisynth.h"


class ChromaShift : public GenericVideoFilter {
	int CShift, UShift, VShift, LShift;
	
public:
	ChromaShift(PClip _child,	int _CShift, int _UShift,
		int _VShift, int _LShift, IScriptEnvironment* env) :
	GenericVideoFilter(_child),CShift(_CShift),UShift(_UShift),VShift(_VShift),LShift(_LShift)
	{
		// parameter checking goes in here
		if (!vi.IsYUY2())
			env->ThrowError("ChromaShift: YUY2 data only (no RGB); use ConvertToYUY2");
		if (((CShift>>1)<<1) != CShift)
			env->ThrowError("ChromaShift: C value must be a mutiple of 2");
		if (((UShift>>1)<<1) != UShift)
			env->ThrowError("ChromaShift: U value must be a mutiple of 2");
		if (((VShift>>1)<<1) != VShift)
			env->ThrowError("ChromaShift: V value must be a mutiple of 2");
	}
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};


PVideoFrame __stdcall ChromaShift::GetFrame(int n, IScriptEnvironment* env) {

	PVideoFrame frame = child->GetFrame(n, env);

   env->MakeWritable(&frame);

   unsigned char* framep = frame->GetWritePtr();
   const int frame_pitch = frame->GetPitch();
   const int frame_row_size = frame->GetRowSize();
   const int frame_height = frame->GetHeight();

	// CShift (Shift both U & V)
	framep = frame->GetWritePtr();
	if (CShift > 0) {
		for (int y = 0; y < frame_height; y++) {
			for (int x = (frame_row_size - 4); x >= (CShift*2); x-=4) {
				framep[x+1] = framep[x+1-(CShift*2)];
				framep[x+3] = framep[x+3-(CShift*2)];
			}
			framep += frame_pitch;
		}
	}
	if (CShift < 0) {
		for (int y = 0; y < frame_height; y++) {
			for (int x = 0; x < (frame_row_size + (CShift*2)); x+=4) {
				framep[x+1] = framep[x+1-(CShift*2)];
				framep[x+3] = framep[x+3-(CShift*2)];
			}
			framep += frame_pitch;
		}
	}

	
	// UShift
	framep = frame->GetWritePtr();
	if (UShift > 0) {
		for (int y = 0; y < frame_height; y++) {
			for (int x = (frame_row_size - 4); x >= (UShift*2); x-=4) {
				framep[x+1] = framep[x+1-(UShift*2)];
			}
			framep += frame_pitch;
		}
	}
	if (UShift < 0) {
		for (int y = 0; y < frame_height; y++) {
			for (int x = 0; x < (frame_row_size + (UShift*2)); x+=4) {
				framep[x+1] = framep[x+1-(UShift*2)];
			}
			framep += frame_pitch;
		}
	}

	

	// VShift
	framep = frame->GetWritePtr();
	if (VShift > 0) {
		for (int y = 0; y < frame_height; y++) {
			for (int x = (frame_row_size - 4); x >= (VShift*2); x-=4) {
				framep[x+3] = framep[x+3-(VShift*2)];
			}
			framep += frame_pitch;;
		}
	}
	if (VShift < 0) {
		for (int y = 0; y < frame_height; y++) {
			for (int x = 0; x < (frame_row_size + (VShift*2)); x+=4) {
				framep[x+3] = framep[x+3-(VShift*2)];
			}
			framep += frame_pitch;
		}
	}

	// LShift
	if (LShift > 0) {
		framep = frame->GetWritePtr();
		framep = framep + (frame_height*frame_pitch) - frame_pitch;

		for (int y = (frame_height - 1); y >= LShift; y--) {
			for (int x = 0; x < frame_row_size; x+=4) {
				framep[x+1] = framep[x+1-(frame_pitch*LShift)];
				framep[x+3] = framep[x+3-(frame_pitch*LShift)];
			}
			framep -= frame_pitch;
		}
	}

	if (LShift < 0) {
		framep = frame->GetWritePtr();
		for (int y = 0; y < (frame_height+ LShift); y++) {
			for (int x = 0; x < frame_row_size; x+=4) {
				framep[x+1] = framep[x+1-(frame_pitch*LShift)];
				framep[x+3] = framep[x+3-(frame_pitch*LShift)];
			}
			framep += frame_pitch;
		}
	}



	return frame;
}


AVSValue __cdecl Create_ChromaShift(AVSValue args, void* user_data, IScriptEnvironment* env) {
    return new ChromaShift(args[0].AsClip(),
			args[1].AsInt(0),
			args[2].AsInt(0),
			args[3].AsInt(0),
			args[4].AsInt(0),
			env);
}


extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit(IScriptEnvironment* env) {
    env->AddFunction("ChromaShift", "c[C]i[U]i[V]i[L]i", Create_ChromaShift, 0);
    return "`ChromaShift' ChromaShift plugin";
}

