/*
    Released under GNU
		You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Simon Walters
	siwalters@hotmail.com

	New versions of the source code and the compiled filter can be 
	found at http://www.geocities.com/siwalters_uk/fnews.html
*/

#include "filter.h"
#include "resource.h"
#include "ScriptError.h"
#include "ScriptInterpreter.h"
#include "ScriptValue.h"
#include "VBitmap.h"
#include "stdio.h"

// Added to use Donald Graft's Dialog Box
#include <windows.h>
#include <commctrl.h>
//#include <crtdbg.h>

//Other includes sometimes needed
#include <string>
//#include <fstream>
//#include <iostream>


///////////////////////
static int RunProc(const FilterActivation *fa, const FilterFunctions *ff);
static long ParamProc(FilterActivation *fa, const FilterFunctions *ff);
static int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
static int InitProc(FilterActivation *fa, const FilterFunctions *ff);
static void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *buf); 
static bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);
static void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc); 


typedef struct MyFilterData {
	IFilterPreview		*ifp;  // Something to do with preview
	int XShift;
	int YShift;
	bool bScroll;
	int XStart;
	int YStart;
	bool bShift;
	bool bLeft;
	bool bDown;
    } MyFilterData;

//////////////////////

static bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	_snprintf(buf, buflen, "Config(%d, %d, %d, %d, %d)",mfd->XShift,mfd->YShift,mfd->bScroll,
					mfd->bLeft,mfd->bDown);

	return true;
}

static void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
	FilterActivation *fa = (FilterActivation *)lpVoid;
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->XShift		= argv[0].asInt();
	mfd->YShift		= argv[1].asInt();
	mfd->bScroll	= !!argv[2].asInt();
	mfd->bLeft		= !!argv[3].asInt();
	mfd->bDown		= !!argv[4].asInt();
//	mfd->bShift		= !!argv[3].asInt();
}

ScriptFunctionDef func_defs[]={
	{ (ScriptFunctionPtr)ScriptConfig, "Config", "0iiiii" },
	{ NULL },
};

CScriptObject script_obj={
	NULL, func_defs
};


///////////////////////////////
struct FilterDefinition filterDef_tutorial = {

    NULL, NULL, NULL,							// next, prev, module
#ifdef _DEBUG
		 "0 ShiftWrap V0.2 beta1 Debug",								// name
#else
		"ShiftWrap V0.2 beta1",								// name
#endif

    "Shift or wraps the frame",				// desc
    "Simon Walters",								// maker
    NULL,											// private_data
    sizeof(MyFilterData),						// inst_data_size

	InitProc,												// initProc
   NULL,												// deinitProc
   RunProc,								// runProc
   ParamProc,												// paramProc
	ConfigProc,												// configProc
   StringProc,							// stringProc
    NULL,											// startProc
    NULL,											// endProc

   &script_obj,											// &script_obj
	FssProc,													// fssProc


};


///////////////////////////////////////////////////////////////


extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_tutorial;
////////////////////////////////////////////////////////////

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_tutorial = ff->addFilter(fm, &filterDef_tutorial, sizeof(FilterDefinition))))
        return 1;

    vdfd_ver    = VIRTUALDUB_FILTERDEF_VERSION;
    vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_tutorial);
}

//////////////////////////////////////

static int RunProc(const FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
   PixDim w, h;
   Pixel32 *src, *dst;
   int lXShift,lYShift;

	if (mfd->bLeft) 
	   lXShift = fa->src.w - mfd->XShift;
	else
		lXShift= mfd->XShift;

	if (mfd->bDown) 
	   lYShift = fa->src.h - mfd->YShift;
	else
		lYShift= mfd->YShift;


	if (fa->pfsi->lCurrentSourceFrame == 0) {
		mfd->XStart=lXShift;
		mfd->YStart=lYShift;
	}
	
	src = (Pixel32 *)((char *)fa->src.data);
	dst = (Pixel32 *)((char *)fa->dst.data);

   h = 0;
	int hcount = mfd->YStart;
	do {
		w = 0;
		int wcount = mfd->XStart;
      do {
			src = (Pixel32 *)((char *)fa->src.data +(fa->src.pitch * h) + (w*4));
			dst = (Pixel32 *)((char *)fa->dst.data + (fa->dst.pitch * hcount) + (wcount*4));
			*dst = *src;
			wcount++;
			if (wcount >= fa->dst.w) {
				wcount = 0;
			}
		} while((++w < fa->dst.w));
		hcount++;
		if (hcount >= fa->dst.h) {
			hcount = 0;
		}
   } while((++h < fa->dst.h));

	if (mfd->bScroll == TRUE) {
		mfd->XStart+=lXShift;
		if (mfd->XStart >= fa->dst.w) {
			mfd->XStart-=fa->dst.w;
		}

		mfd->YStart+=lYShift;
		if (mfd->YStart >= fa->dst.h) {
			mfd->YStart-=fa->dst.h;
		}
	}


	return 0;
}


static long ParamProc(FilterActivation *fa, const FilterFunctions *ff) {
    return   FILTERPARAM_SWAP_BUFFERS;
}


BOOL CALLBACK ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
    MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);
		
	 HWND hWnd;
			switch(msg) {
        case WM_INITDIALOG:
            SetWindowLong(hdlg, DWL_USER, lParam);
            mfd = (MyFilterData *)lParam;

            CheckDlgButton(hdlg, IDC_CHECK1, mfd->bScroll?BST_CHECKED:BST_UNCHECKED);
            CheckDlgButton(hdlg, IDC_CHECK2, mfd->bShift?BST_CHECKED:BST_UNCHECKED);
				SendMessage(GetDlgItem(hdlg, IDC_RADIO2), BM_CLICK, 0, 0);
				SendMessage(GetDlgItem(hdlg, IDC_RADIO3), BM_CLICK, 0, 0);


// Must set Edit box to be 1 before spin control in TAB order
// Also set auto buddy and Set Buddy Integer
				SetDlgItemInt(hdlg, IDC_EDIT1, mfd->XShift, FALSE);
				hWnd = GetDlgItem(hdlg, IDC_SPIN1);
				SendMessage(hWnd, UDM_SETRANGE, (WPARAM)TRUE, MAKELONG((352 - 1),0));
				SendMessage(hWnd, UDM_SETPOS, (WPARAM)TRUE, mfd->XShift);
	
				SetDlgItemInt(hdlg, IDC_EDIT2, mfd->YShift, FALSE);
				hWnd = GetDlgItem(hdlg, IDC_SPIN2);
				SendMessage(hWnd, UDM_SETRANGE, (WPARAM)TRUE, MAKELONG((576 - 1),0));
				SendMessage(hWnd, UDM_SETPOS, (WPARAM)TRUE, mfd->YShift);

				mfd->XStart=mfd->XShift;
				mfd->YStart=mfd->YShift;
            return TRUE;

			case WM_VSCROLL:
				if ((HWND) lParam == GetDlgItem(hdlg, IDC_SPIN1))
				{
					int XShift = SendMessage(GetDlgItem(hdlg, IDC_SPIN1), UDM_GETPOS, 0, 0);
					if (XShift != mfd->XShift)
					{
						mfd->XShift = XShift;
						mfd->XStart = XShift;
						SendMessage(GetDlgItem(hdlg, IDC_EDIT1), TBM_SETPOS, (WPARAM)TRUE, (mfd->XShift));
					}
				}
				else if ((HWND) lParam == GetDlgItem(hdlg, IDC_SPIN2))
				{
					int YShift = SendMessage(GetDlgItem(hdlg, IDC_SPIN2), UDM_GETPOS, 0, 0);
					if (YShift != mfd->YShift)
					{
						mfd->YShift = YShift;
						mfd->YStart = YShift;
						SendMessage(GetDlgItem(hdlg, IDC_EDIT2), TBM_SETPOS, (WPARAM)TRUE, (mfd->YShift));
					}
				}

			break;
			case WM_COMMAND:
 
				switch(LOWORD(wParam)) {

				case IDOK:
					mfd->bScroll = !!IsDlgButtonChecked(hdlg, IDC_CHECK1);
					mfd->bShift = !!IsDlgButtonChecked(hdlg, IDC_CHECK2);
					mfd->bLeft = !!IsDlgButtonChecked(hdlg, IDC_RADIO1);
					mfd->bDown = !!IsDlgButtonChecked(hdlg, IDC_RADIO4);

					EndDialog(hdlg, 0);
					return TRUE;
				case IDCANCEL:
					EndDialog(hdlg, 1);
					return FALSE;
				}
            break;
    }

    return FALSE;
}

static int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {
    return DialogBoxParam(fa->filter->module->hInstModule,
            MAKEINTRESOURCE(IDD_FILTER), hwnd,
            (DLGPROC)ConfigDlgProc, (LPARAM)fa->filter_data);
}

// Set any parameters to a default value
int InitProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->XShift	= 0;
	mfd->YShift	= 0;
	mfd->XStart=0;
	mfd->YStart=0;
	mfd->bScroll=FALSE;
	mfd->bShift=TRUE;
	mfd->bLeft=FALSE;
	mfd->bDown=FALSE;
	return 0;
}

// Displays the parameters in the total filters loaded view
void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {

    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	
	// write the parameters to a string (YT: xxx, YB: xxx)
	sprintf(str, "(XS: %d, YS: %d, Scrl: %d)", mfd->XShift, mfd->YShift,
		mfd->bScroll);
}


/*
			case IDC_RADIO21:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO31:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO41:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO51:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO61:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO71:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO81:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO10:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO20:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO30:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO40:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO50:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO60:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO70:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO80:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO1:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO2:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO3:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO4:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO5:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO6:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO7:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			case IDC_RADIO8:
				ReadDialogBits(&mfd->Mask0, &mfd->Mask1, hdlg);
				mfd->ifp->RedoFrame();
				break;
			}
			break;
	}

	return FALSE;
}


static int tutorialConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd)
{
	MyFilterData *mfd = (MyFilterData *) fa->filter_data;
	MyFilterData mfd_old = *mfd;
	int ret;

	mfd->ifp = fa->ifp;
	if (DialogBoxParam(fa->filter->module->hInstModule,
			MAKEINTRESOURCE(IDD_FILTER), hwnd,
			tutorialConfigDlgProc, (LPARAM) mfd))
	{
		*mfd = mfd_old;
		ret = TRUE;
	}
	else
	{
		ret = FALSE;
	}
	return(ret);
}




static void tutorialStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {

    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	
	// write the parameters to a string 
	sprintf(str, "(0 Mask: %d, 1 Mask: %d )",  mfd->Mask0,mfd->Mask1);
}

//Script/Batch Handling Code


static bool tutorialFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	// Write the two parameters to a string 
    _snprintf(buf, buflen, "Config(%d %d)",
        mfd->Mask0, mfd->Mask1);

    return true;
}


static void tutorialScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
    FilterActivation *fa = (FilterActivation *)lpVoid;
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	// Extract the parameters from the arguments provided
    mfd->Mask0   = argv[0].asInt();
    mfd->Mask1   = argv[1].asInt();
}

static void SetDialogBits(unsigned char Mask0, unsigned char Mask1, HWND hdlg) {
	SendMessage(GetDlgItem(hdlg, IDC_RADIO1), BM_CLICK, 0, 0);
	SendMessage(GetDlgItem(hdlg, IDC_RADIO2), BM_CLICK, 0, 0);
	SendMessage(GetDlgItem(hdlg, IDC_RADIO3), BM_CLICK, 0, 0);
	SendMessage(GetDlgItem(hdlg, IDC_RADIO4), BM_CLICK, 0, 0);
	SendMessage(GetDlgItem(hdlg, IDC_RADIO5), BM_CLICK, 0, 0);
	SendMessage(GetDlgItem(hdlg, IDC_RADIO6), BM_CLICK, 0, 0);
	SendMessage(GetDlgItem(hdlg, IDC_RADIO7), BM_CLICK, 0, 0);
	SendMessage(GetDlgItem(hdlg, IDC_RADIO8), BM_CLICK, 0, 0);
	if (Mask1 & 1) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO11);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
		else if ((Mask0 & 1)==0) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO10);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	if (Mask1 & 2) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO21);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	else if ((Mask0 & 2)==0) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO20);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	if (Mask1 & 4) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO31);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	else if ((Mask0 & 4)==0) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO30);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	if (Mask1 & 8) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO41);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	else if ((Mask0 & 8)==0) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO40);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	if (Mask1 & 16) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO51);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	else if ((Mask0 & 16)==0) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO50);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	if (Mask1 & 32) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO61);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	else if ((Mask0 & 32)==0) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO60);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	if (Mask1 & 64) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO71);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	else if ((Mask0 & 64)==0) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO70);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	if (Mask1 & 128) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO81);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	else if ((Mask0 & 128)==0) {
		HWND hWnd = GetDlgItem(hdlg, IDC_RADIO80);
		SendMessage(hWnd, BM_CLICK, 0, 0);
	}
	//

}


static void ReadDialogBits(int *Mask0, int *Mask1, HWND hdlg)	{
	*Mask0 = 255;
	*Mask1 = 0;

	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO11), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask1 = *Mask1 + 1;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO21), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask1 = *Mask1 + 2;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO31), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask1 = *Mask1 + 4;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO41), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask1 = *Mask1 + 8;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO51), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask1 = *Mask1 + 16;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO61), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask1 = *Mask1 + 32;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO71), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask1 = *Mask1 + 64;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO81), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask1 = *Mask1 + 128;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO10), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask0 = *Mask0 - 1;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO20), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask0 = *Mask0 - 2;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO30), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask0 = *Mask0 - 4;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO40), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask0 = *Mask0 - 8;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO50), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask0 = *Mask0 - 16;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO60), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask0 = *Mask0 - 32;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO70), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask0 = *Mask0 - 64;

		}
	if (SendMessage(GetDlgItem(hdlg, IDC_RADIO80), BM_GETCHECK, 0,0) == BST_CHECKED) {
		*Mask0 = *Mask0 - 128;
		}
		SetDlgItemInt(hdlg, IDC_EDIT_MASK0,*Mask0,FALSE);
		SetDlgItemInt(hdlg, IDC_EDIT_MASK1,*Mask1,FALSE);
}


*/
