#include <windows.h>
#include <stdio.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "filter.h"


int RunProc(const FilterActivation *fa, const FilterFunctions *ff);
long ParamProc(FilterActivation *fa, const FilterFunctions *ff);


struct FilterDefinition filterDef = {

    NULL, NULL, NULL,       // next, prev, module
    "AnamorphicResize",     // name
    "This filter reduces the vertical image size by 25% using the same algorithm as DVD-players !\nVersion 0.10 (Mar/30/2001)",
                            // desc
    "tHE fISH",             // maker
    NULL,                   // private_data
    0,                      // inst_data_size

    NULL,                   // initProc
    NULL,                   // deinitProc
    RunProc,                // runProc
    ParamProc,              // paramProc
    NULL,                   // configProc
    NULL,                   // stringProc
    NULL,                   // startProc
    NULL,                   // endProc

    NULL,                   // script_obj
    NULL,                   // fssProc

};


extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd;

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd = ff->addFilter(fm, &filterDef, sizeof(FilterDefinition))))
        return 1;

    vdfd_ver    = VIRTUALDUB_FILTERDEF_VERSION;
    vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd);
}


int RunProc(const FilterActivation *fa, const FilterFunctions *ff) {
    int w, h;
    Pixel32 *src, *dst;
	int i;

    src = (Pixel32 *)fa->src.data;
    dst = (Pixel32 *)fa->dst.data;

	h = fa->src.h;
	w = fa->src.w;

	for (i=0; i<h; i+=4) {
		Pixel32 *prow1 = ((Pixel32 *)((char *)src));
        Pixel32 *prow2 = ((Pixel32 *)((char *)src + fa->src.pitch));
        Pixel32 *prow3 = ((Pixel32 *)((char *)src + (2*fa->src.pitch)));
        Pixel32 *prow4 = ((Pixel32 *)((char *)src + (3*fa->src.pitch)));
  	    
		for(int x = 0; x < w; x++) {
			int row1 = *prow1++;
			int row2 = *prow2++;
			int row3 = *prow3++;
			int row4 = *prow4++;

			int r1R = ((row1 >> 16) & 0xff);
			int r1G = ((row1 >> 8) & 0xff);
			int r1B = (row1 & 0xff);
			int r2R = ((row2 >> 16) & 0xff);
			int r2G = ((row2 >> 8) & 0xff);
			int r2B = (row2 & 0xff);
			int r3R = ((row3 >> 16) & 0xff);
			int r3G = ((row3 >> 8) & 0xff);
			int r3B = (row3 & 0xff);
			int r4R = ((row4 >> 16) & 0xff);
			int r4G = ((row4 >> 8) & 0xff);
			int r4B = (row4 & 0xff);

			double med1 = (((0.75*r1B) + (0.25*r2B)) +
				            ((int)((0.75*r1G) + (int)(0.25*r2G)) << 8) +
				            ((int)((0.75*r1R) + (int)(0.25*r2R)) << 16) );
			double med2 = (((0.5*r2B) + (0.5*r3B)) +
				            ((int)((0.5*r2G) + (int)(0.5*r3G)) << 8) +
				            ((int)((0.5*r2R) + (int)(0.5*r3R)) << 16) );
			double med3 = (((0.25*r3B) + (0.75*r4B)) +
				            ((int)((0.25*r3G) + (int)(0.75*r4G)) << 8) +
				            ((int)((0.25*r3R) + (int)(0.75*r4R)) << 16) );

            *(dst+x) = (Pixel32) med1;
			*(dst+x+w) = (Pixel32) med2;
			*(dst+x+(2*w)) = (Pixel32) med3;

		}

		src = (Pixel32 *)((char *)src + (4 * fa->src.pitch));
		dst = (Pixel32 *)((char *)dst + (3 * fa->dst.pitch));

	}

	return 0;
}





long ParamProc(FilterActivation *fa, const FilterFunctions *ff) {

	fa->dst.h *= 0.75;

    return 0;
}

