/*
    Released under GNU
		You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Simon Walters
	siwalters@hotmail.com

	New versions of the source code and the compiled filter can be 
	found at http://www.geocities.com/siwalters_uk/fnews.html
*/

#include "filter.h"

///////////////////////
static int UnViewFieldsRunProc(const FilterActivation *fa, const FilterFunctions *ff);
static long UnViewFieldsParamProc(FilterActivation *fa, const FilterFunctions *ff);


///////////////////////////////
struct FilterDefinition filterDef_tutorial = {

    NULL, NULL, NULL,							// next, prev, module
#ifdef _DEBUG
		 "0 UnViewFields V1.2 Dbg",								// name
#else
		"UnViewFields V1.2",								// name
#endif

    "Lets you unview viewed fields",				// desc
    "Simon Walters",								// maker
    NULL,											// private_data
    NULL,						// inst_data_size

	NULL,												// initProc
   NULL,												// deinitProc
   UnViewFieldsRunProc,								// runProc
   UnViewFieldsParamProc,												// paramProc
	NULL,												// configProc
   NULL,							// stringProc
    NULL,											// startProc
    NULL,											// endProc

    NULL,								// script_obj
    NULL,								// fssProc


};


///////////////////////////////////////////////////////////////


extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_tutorial;
////////////////////////////////////////////////////////////

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_tutorial = ff->addFilter(fm, &filterDef_tutorial, sizeof(FilterDefinition))))
        return 1;

    vdfd_ver    = VIRTUALDUB_FILTERDEF_VERSION;
    vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_tutorial);
}

//////////////////////////////////////

static int UnViewFieldsRunProc(const FilterActivation *fa, const FilterFunctions *ff) {

   Pixel32 *srcp, *dstp, *srcp2;

	srcp = (Pixel32 *)fa->src.data;
	dstp = (Pixel32 *)fa->dst.data;

    const int dst_pitch = (fa->src.pitch / 4);
    const int dst_row_size = fa->src.w * 4;
    const int dst_height = fa->src.h;

	srcp2 = srcp + (dst_height >> 1) * dst_pitch;

   int h;

	h = dst_height >> 1;
	do {

		memcpy(dstp,srcp,dst_row_size);

		srcp = srcp + dst_pitch;
		dstp = dstp + dst_pitch;

		memcpy(dstp,srcp2,dst_row_size);
		srcp2 = srcp2 + dst_pitch;
		dstp = dstp + dst_pitch;

   } while(--h);


	return 0;
}






static long UnViewFieldsParamProc(FilterActivation *fa, const FilterFunctions *ff) {
    return  FILTERPARAM_SWAP_BUFFERS;

}



