//	filtfactory - poor man's filter compiler for VirtualDub
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef f_FILTFACTORY_CODEGEN_H
#define f_FILTFACTORY_CODEGEN_H

class ParseNodeAllocator;
class TreeTranslationNode;

enum InstructionType {

	// Flow control

	INST_JMP,					// pc <- #r2
	INST_JNZ,					// pc <- r1 ? #r2 : pc
	INST_JZ,					// pc <- r1 ? pc : #r2
	INST_CALL,					// *sp++ = fp, *sp++ = pc, pc = #r2
	INST_RVALINT,				// r0 <- rv
	INST_RVALBITMAP,			// r0 <- rv
	INST_RVALPIXEL32,			// r0 <- rv
	INST_RVALPIXEL64,			// r0 <- rv
	INST_RET,					// pc = *--sp; fp = *--sp; sp -= #r2
	INST_RETINT,				// rv <- r1; pc = *--sp; fp = *--sp; sp -= #r2
	INST_RETBITMAP,				// rv <- r1; pc = *--sp; fp = *--sp; sp -= #r2
	INST_RETPIXEL32,			// rv <- r1; pc = *--sp; fp = *--sp; sp -= #r2
	INST_RETPIXEL64,			// rv <- r1; pc = *--sp; fp = *--sp; sp -= #r2

	// Stack control

	INST_LINK,					// assert if #r1 dwords not available
	INST_PUSHINT,				// *sp++ = r1
	INST_PUSHBITMAP,			// *sp++ = r1
	INST_PUSHPIXEL32,			// *sp++ = r1
	INST_PUSHPIXEL64,			// *sp++ = r1

	// Bitmap instructions

	INST_MOVEBITMAP,			// r0 <- r1
	INST_GETWBITMAP,			// r0 <- r1.w
	INST_GETHBITMAP,			// r0 <- r1.h
	INST_GETPIXELBITMAP,		// r0 <- r0[r1,r2]
	INST_SETPIXELBITMAP,		// r0[r2,r2+] <- r1
	INST_SCANSTART1,			// r0+1 <- 0, r0+2 <- h, r0+3 <- w, r0+4 <- ptr, r0+5 <- modulo
	INST_SCANSTART2,			// r0+1 <- 0, r0+2 <- h, r0+3 <- w, r0+4 <- ptr1, r0+5 <- ptr2, r0+6 <- modulo1, r0+7 <- modulo2
	INST_SCANEND1,				// pc = ++(r0+4),(++(r0+1) < r1.w) || ((r0+4) += (r0+5)),((r0+1)=(r0+3)) || --(r0+2) ? r1 : pc;
	INST_SCANEND2,				// pc = ++(r0+4),++(r0+5),(++(r0+1) < r1.w) || ((r0+4) += (r0+6)),((r0+5) += (r0+7)),((r0+1)=(r0+3)) || --(r0+2) ? r1 : pc;

	// Pixel32 instructions

	INST_CONSTPIXEL32,			// r0 <- #r1
	INST_MOVEPIXEL32,			// r0 <- r1
	INST_INVPIXEL32,			// r0 <- ~r1
	INST_ADDPIXEL32,			// r0 <- r1 + r2
	INST_SUBPIXEL32,			// r0 <- r1 - r2
	INST_MULPIXEL32,			// r0 <- r1 * r2
	INST_ANDPIXEL32,			// r0 <- r1 & r2
	INST_ORPIXEL32,				// r0 <- r1 | r2
	INST_XORPIXEL32,			// r0 <- r1 ^ r2
	INST_GETRPIXEL32,			// r0 <- r1.r
	INST_GETGPIXEL32,			// r0 <- r1.g
	INST_GETBPIXEL32,			// r0 <- r1.b
	INST_SETRPIXEL32,			// r0.r <- r1
	INST_SETGPIXEL32,			// r0.g <- r1
	INST_SETBPIXEL32,			// r0.b <- r1
	INST_PIXEL64TOPIXEL32,		// r0 <- r1

	// Pixel64 instructions

	INST_MOVEPIXEL64,			// r0 <- r1
	INST_INVPIXEL64,
	INST_ADDPIXEL64,			// r0 <- r1 + r2
	INST_SUBPIXEL64,			// r0 <- r1 - r2
	INST_MULPIXEL64,			// r0 <- r1 * r2
	INST_ANDPIXEL64,			// r0 <- r1 & r2
	INST_ORPIXEL64,				// r0 <- r1 | r2
	INST_XORPIXEL64,			// r0 <- r1 ^ r2
	INST_LSHIFTPIXEL64,			// r0 <- r1 << r2
	INST_RSHIFTPIXEL64,			// r0 <- r1 >> r2
	INST_RLSHIFTPIXEL64,		// r0 <- r1 >>> r2
	INST_GETRPIXEL64,			// r0 <- r1.r
	INST_GETGPIXEL64,			// r0 <- r1.g
	INST_GETBPIXEL64,			// r0 <- r1.b
	INST_SETRPIXEL64,			// r0.r <- r1
	INST_SETGPIXEL64,			// r0.g <- r1
	INST_SETBPIXEL64,			// r0.b <- r1
	INST_PIXEL32TOPIXEL64,		// r0 <- r1

	// Integer instructions

	INST_CONSTINT,				// r0 <- #r1
	INST_MOVEINT,				// r0 <- r1
	INST_NEGINT,				// r0 <- -r1
	INST_INVINT,				// r0 <- ~r1
	INST_INCINT,				// ++r0;
	INST_DECINT,				// --r0;
	INST_ADDINT,				// r0 <- r1 + r2
	INST_ADDCONSTINT,			// r0 <- r1 + #r2
	INST_SUBINT,				// r0 <- r1 - r2
	INST_MULINT,				// r0 <- r1 * r2
	INST_DIVINT,				// r0 <- r1 / r2
	INST_ANDINT,				// r0 <- r1 & r2
	INST_ORINT,					// r0 <- r1 | r2
	INST_XORINT,				// r0 <- r1 ^ r2
	INST_LSHIFTINT,				// r0 <- r1 << r2
	INST_RSHIFTINT,				// r0 <- r1 >> r2
	INST_RLSHIFTINT,			// r0 <- r1 >>> r2
	INST_LANDINT,				// r0 <- r1 && r2
	INST_LORINT,				// r0 <- r1 || r2
	INST_CMPLTINT,				// r0 <- r1 < r2
	INST_CMPLEINT,				// r0 <- r1 <= r2
	INST_CMPGTINT,				// r0 <- r1 > r2
	INST_CMPGEINT,				// r0 <- r1 >= r2
	INST_CMPEQINT,				// r0 <- r1 == r2
	INST_CMPNEINT,				// r0 <- r1 != r2
};

struct Instruction {
	InstructionType	opcode;
	int	r0, r1, r2;
};

class CodeGenerator {
private:
	ParseNodeAllocator *const autofree;
	TreeTranslationNode *const pUnit;

public:
	CodeGenerator(ParseNodeAllocator *autofree, TreeTranslationNode *pUnit);
	~CodeGenerator();

	Instruction *Compile();
};

#endif
