//	filtfactory - poor man's filter compiler for VirtualDub
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <crtdbg.h>
#include <stdio.h>

#include "filter.h"
#include "codegen.h"

static const struct {
	char *name;
	int type;
} inst_lookup[]={
	"JMP",					2,
	"JNZ",					6,
	"JZ",					6,
	"CALL",					7,
	"RVALINT",				5,
	"RVALBITMAP",			5,
	"RVALPIXEL32",			5,
	"RVALPIXEL64",			5,
	"RET",					2,
	"RETINT",				1,
	"RETBITMAP",			1,
	"RETPIXEL32",			1,
	"RETPIXEL64",			1,
	"LINK",					7,
	"PUSHINT",				8,
	"PUSHBITMAP",			8,
	"PUSHPIXEL32",			8,
	"PUSHPIXEL64",			8,
	"MOVEBITMAP",			5,
	"GETWBITMAP",			1,
	"GETHBITMAP",			1,
	"GETPIXELBITMAP",		1,
	"SETPIXELBITMAP",		1,
	"SCANSTART1",			5,
	"SCANSTART2",			1,
	"SCANEND1",				6,
	"SCANEND2",				6,
	"CONSTPIXEL32",			4,
	"MOVEPIXEL32",			5,
	"INVPIXEL32",			5,
	"ADDPIXEL32",			1,
	"SUBPIXEL32",			1,
	"MULPIXEL32",			1,
	"ANDPIXEL32",			1,
	"ORPIXEL32",			1,
	"XORPIXEL32",			1,
	"GETRPIXEL32",			1,
	"GETGPIXEL32",			1,
	"GETBPIXEL32",			1,
	"SETRPIXEL32",			1,
	"SETGPIXEL32",			1,
	"SETBPIXEL32",			1,
	"PIXEL64TOPIXEL32",		1,
	"MOVEPIXEL64",			5,
	"INVPIXEL64",			5,
	"ADDPIXEL64",			1,
	"SUBPIXEL64",			1,
	"MULPIXEL64",			1,
	"ANDPIXEL64",			1,
	"ORPIXEL64",			1,
	"XORPIXEL64",			1,
	"LSHIFTPIXEL64",		1,
	"RSHIFTPIXEL64",		1,
	"RLSHIFTPIXEL64",		1,
	"GETRPIXEL64",			1,
	"GETGPIXEL64",			1,
	"GETBPIXEL64",			1,
	"SETRPIXEL64",			1,
	"SETGPIXEL64",			1,
	"SETBPIXEL64",			1,
	"PIXEL32TOPIXEL64",		1,
	"CONSTINT",				3,
	"MOVEINT",				5,
	"NEGINT",				5,
	"INVINT",				5,
	"INCINT",				5,
	"DECINT",				5,
	"ADDINT",				1,
	"ADDCONSTINT",			1,
	"SUBINT",				1,
	"MULINT",				1,
	"DIVINT",				1,
	"ANDINT",				1,
	"ORINT",				1,
	"XORINT",				1,
	"LSHIFTINT",			1,
	"RSHIFTINT",			1,
	"RLSHIFTINT",			1,
	"LANDINT",				1,
	"LORINT",				1,
	"CMPLTINT",				1,
	"CMPLEINT",				1,
	"CMPGTINT",				1,
	"CMPGEINT",				1,
	"CMPEQINT",				1,
	"CMPNEINT",				1,
};

void Disassemble(int i, const Instruction& inst) {
	char buf[256];

	switch(inst_lookup[(int)inst.opcode].type) {
	case 0:
		sprintf(buf, "%4d: %15s\n", i, inst_lookup[(int)inst.opcode].name);
		break;
	case 1:
		sprintf(buf, "%4d: %15s r%d, r%d, r%d\n", i, inst_lookup[(int)inst.opcode].name, inst.r0, inst.r1, inst.r2);
		break;
	case 2:
		sprintf(buf, "%4d: %15s #%d\n", i, inst_lookup[(int)inst.opcode].name, inst.r2);
		break;
	case 3:
		sprintf(buf, "%4d: %15s r%d, %d\n", i, inst_lookup[(int)inst.opcode].name, inst.r0, inst.r1);
		break;
	case 4:
		sprintf(buf, "%4d: %15s r%d, #%6x\n", i, inst_lookup[(int)inst.opcode].name, inst.r0, inst.r1);
		break;
	case 5:
		sprintf(buf, "%4d: %15s r%d, r%d\n", i, inst_lookup[(int)inst.opcode].name, inst.r0, inst.r1);
		break;
	case 6:
		sprintf(buf, "%4d: %15s r%d, #%d\n", i, inst_lookup[(int)inst.opcode].name, inst.r1, inst.r2);
		break;
	case 7:
		sprintf(buf, "%4d: %15s #%d, #%d\n", i, inst_lookup[(int)inst.opcode].name, inst.r1, inst.r2);
		break;
	case 8:
		sprintf(buf, "%4d: %15s r%d\n", i, inst_lookup[(int)inst.opcode].name, inst.r1);
		break;
	case 9:
		sprintf(buf, "%4d: %15s r%d, #%d, #%d\n", i, inst_lookup[(int)inst.opcode].name, inst.r0, inst.r1, inst.r2);
		break;
	}
	_RPT0(0,buf);
}

void InterpretCode(Instruction *code, int pc, long *stack, int sp, long& r1, long& r2) {
	long return1, return2;
	int fp = -1;

	for(;;) {
		const Instruction& inst = code[pc++];
		long u, v;

//		_RPT4(0,"r0=% 8lx, r1=% 8lx, r2=% 8lx, r3=% 8lx\n", stack[fp+0], stack[fp+1], stack[fp+2], stack[fp+3]);

//		Disassemble(pc-1, inst);

		switch(inst.opcode) {
		case INST_JMP:					// pc <- #r2
			pc = inst.r2;
			break;
		case INST_JNZ:					// pc <- r1 ? #r2 : pc
			if (stack[fp+inst.r1])
				pc = inst.r2;
			break;
		case INST_JZ:					// pc <- r1 ? pc : #r2
			if (!stack[fp+inst.r1])
				pc = inst.r2;
			break;
		case INST_CALL:					// *sp++ = fp, *sp++ = pc, pc = #r2
			stack[sp++] = fp;
			stack[sp++] = pc;
			pc = inst.r2;
			break;
		case INST_RVALINT:
			stack[fp+inst.r0] = return1;
			break;
		case INST_RVALBITMAP:
			stack[fp+inst.r0] = return1;
			break;
		case INST_RVALPIXEL32:
			stack[fp+inst.r0] = return1;
			break;
		case INST_RVALPIXEL64:
			stack[fp+inst.r0] = return1;
			stack[fp+inst.r0+1] = return2;
			break;
		case INST_RET:					// pc = *--sp; fp = *--sp; sp -= #r2
			sp = fp;
			pc = stack[--sp];
			fp = stack[--sp];
			sp -= inst.r2;

			if (fp<0)
				goto finished;
			break;
		case INST_RETINT:				// pc = *--sp; fp = *--sp; sp -= #r2
			return1 = stack[fp+inst.r1];
			sp = fp;
			pc = stack[--sp];
			fp = stack[--sp];
			sp -= inst.r2;

			if (fp<0)
				goto finished;
			break;
		case INST_RETBITMAP:			// pc = *--sp; fp = *--sp; sp -= #r2
			return1 = stack[fp+inst.r1];
			sp = fp;
			pc = stack[--sp];
			fp = stack[--sp];
			sp -= inst.r2;

			if (fp<0)
				goto finished;
			break;
		case INST_RETPIXEL32:			// pc = *--sp; fp = *--sp; sp -= #r2
			return1 = stack[fp+inst.r1];
			sp = fp;
			pc = stack[--sp];
			fp = stack[--sp];
			sp -= inst.r2;

			if (fp<0)
				goto finished;
			break;
		case INST_RETPIXEL64:			// pc = *--sp; fp = *--sp; sp -= #r2
			return1 = stack[fp+inst.r1];
			return2 = stack[fp+inst.r1+1];
			sp = fp;
			pc = stack[--sp];
			fp = stack[--sp];
			sp -= inst.r2;

			if (fp<0)
				goto finished;
			break;

		// Stack control

		case INST_LINK:					// assert if #r2 dwords not available
			fp = sp;
			sp += inst.r2;
			break;
		case INST_PUSHINT:				// *sp++ = r1
			stack[sp++] = stack[fp+inst.r1];
			break;
		case INST_PUSHBITMAP:			// *sp++ = r1
			stack[sp++] = stack[fp+inst.r1];
			break;
		case INST_PUSHPIXEL32:			// *sp++ = r1
			stack[sp++] = stack[fp+inst.r1];
			break;
		case INST_PUSHPIXEL64:			// *sp++ = r1
			stack[sp+0] = stack[fp+inst.r1+0];
			stack[sp+1] = stack[fp+inst.r1+1];
			sp += 2;
			break;

		// Bitmap instructions

		case INST_MOVEBITMAP:			// r0 <- r1
			stack[fp+inst.r0] = stack[fp+inst.r1];
			break;
		case INST_GETWBITMAP:			// r0 <- r1.w
			stack[fp+inst.r0] = ((VBitmap *)stack[fp+inst.r1])->w;
			break;
		case INST_GETHBITMAP:			// r0 <- r1.h
			stack[fp+inst.r0] = ((VBitmap *)stack[fp+inst.r1])->h;
			break;
		case INST_GETPIXELBITMAP:		// r0 <- r0[r1,r2]
			{
				int x = stack[fp+inst.r1];
				int y = stack[fp+inst.r2];
				const VBitmap *pvbm = (VBitmap *)stack[fp+inst.r0];

				if (x < 0)
					x = 0;
				
				if (y < 0)
					y=0;
				
				if (x >= pvbm->w)
					x = pvbm->w-1;
				
				if (y >= pvbm->h)
					y = pvbm->h-1;

				stack[fp+inst.r0] = *pvbm->Address32(x, y);
			}
			break;
		case INST_SETPIXELBITMAP:		// r0[r2,r2+] <- r1
			{
				int x = stack[fp+inst.r2+0];
				int y = stack[fp+inst.r2+1];
				const VBitmap *pvbm = (VBitmap *)stack[fp+inst.r0];

				if ((x|y) >= 0 && x < pvbm->w && y < pvbm->h)
					*pvbm->Address32(x, y) = stack[fp+inst.r1];
			}
			break;

		case INST_SCANSTART1:
			{
				long *frame = stack + fp + inst.r0;
				VBitmap *dst = (VBitmap *)stack[fp+inst.r1];

				frame[0] = dst->data[0];
				frame[1] = 0;
				frame[2] = dst->h;
				frame[3] = dst->w;
				frame[4] = (long)dst->data;
				frame[5] = (long)dst->modulo;
			}
			break;

		case INST_SCANSTART2:
			{
				long *frame = stack + fp + inst.r0;
				VBitmap *dst = (VBitmap *)stack[fp+inst.r1];
				VBitmap *src = (VBitmap *)stack[fp+inst.r2];
				int w = src->w < dst->w ? src->w : dst->w;
				int h = src->h < dst->h ? src->h : dst->h;

				frame[0] = src->data[0];
				frame[1] = w;
				frame[2] = h;
				frame[3] = w;
				frame[4] = (long)dst->data;
				frame[5] = (long)src->data;
				frame[6] = dst->pitch - 4*w;
				frame[7] = src->pitch - 4*w;
			}
			break;

		case INST_SCANEND1:
			{
				long *frame = stack + fp + inst.r1;

				*(Pixel32 *)frame[4] = frame[0];
				frame[4] += 4;
				if (!--frame[1]) {
					frame[1] = frame[3];
					frame[4] += frame[5];

					if (!--frame[2])
						break;
				}

				frame[0] = *(Pixel32 *)frame[4];

				pc = inst.r2;
			}
			break;

		case INST_SCANEND2:
			{
				long *frame = stack + fp + inst.r1;

				*(Pixel32 *)frame[4] = frame[0];
				frame[4] += 4;
				frame[5] += 4;
				if (!--frame[1]) {
					frame[1] = frame[3];
					frame[4] += frame[6];
					frame[5] += frame[7];

					if (!--frame[2])
						break;
				}

				frame[0] = *(Pixel32 *)frame[5];
				pc = inst.r2;
			}
			break;

		// Pixel32 instructions

		case INST_CONSTPIXEL32:			// r0 <- #r1
			stack[fp+inst.r0] = inst.r1;
			break;
		case INST_MOVEPIXEL32:			// r0 <- r1
			stack[fp+inst.r0] = stack[fp+inst.r1];
			break;
		case INST_INVPIXEL32:			// r0 <- ~r1
			stack[fp+inst.r0] = ~stack[fp+inst.r1];
			break;
		case INST_ADDPIXEL32:		// r0 <- r1 + r2
			u = stack[fp+inst.r1];
			stack[fp+inst.r0] = ((u & 0x7f7f7f7f) + stack[fp+inst.r2]) ^ (u & 0x80808080);
			break;
		case INST_SUBPIXEL32:			// r0 <- r1 - r2
			u = stack[fp+inst.r1];
			stack[fp+inst.r0] = ((u | 0x80808080) - stack[fp+inst.r2]) ^ (u & 0x80808080);
			break;
		case INST_MULPIXEL32:			// r0 <- r1 * r2
			u = stack[fp+inst.r1];
			v = stack[fp+inst.r2];
			stack[fp+inst.r0]	= ((((u>>16)&0xff) * ((v>>16)&0xff)) << 16)
								+ ((((u>> 8)&0xff) * ((v>> 8)&0xff)) <<  8)
								+ ((((u    )&0xff) * ((v    )&0xff))      );
			break;
		case INST_ANDPIXEL32:
			stack[fp+inst.r0] = stack[fp+inst.r1] & stack[fp+inst.r2];
			break;
		case INST_ORPIXEL32:
			stack[fp+inst.r0] = stack[fp+inst.r1] | stack[fp+inst.r2];
			break;
		case INST_XORPIXEL32:
			stack[fp+inst.r0] = stack[fp+inst.r1] ^ stack[fp+inst.r2];
			break;
		case INST_GETRPIXEL32:			// r0 <- r1.r
			stack[fp+inst.r0] = (stack[fp+inst.r1] >> 16)&0xff;
			break;
		case INST_GETGPIXEL32:			// r0 <- r1.g
			stack[fp+inst.r0] = (stack[fp+inst.r1] >>  8)&0xff;
			break;
		case INST_GETBPIXEL32:			// r0 <- r1.b
			stack[fp+inst.r0] = stack[fp+inst.r1]&0xff;
			break;
		case INST_SETRPIXEL32:			// r0.r <- r1
			stack[fp+inst.r0] = (stack[fp+inst.r0] & 0xff00ffff) + ((stack[fp+inst.r1]&0xff)<<16);
			break;
		case INST_SETGPIXEL32:			// r0.g <- r1
			stack[fp+inst.r0] = (stack[fp+inst.r0] & 0xffff00ff) + ((stack[fp+inst.r1]&0xff)<<8);
			break;
		case INST_SETBPIXEL32:			// r0.b <- r1
			stack[fp+inst.r0] = (stack[fp+inst.r0] & 0xffffff00) + (stack[fp+inst.r1]&0xff);
			break;
		case INST_PIXEL64TOPIXEL32:		// r0 <- r1
			u = stack[fp+inst.r1+1];
			v = stack[fp+inst.r1];

			stack[fp+inst.r0]	= ((u&0x00ff0000) << 8)
								+ ((u&0x000000ff) << 16)
								+ ((v&0x00ff0000) >> 8)
								+ (v&0x000000ff);
			break;

		// Pixel64 instructions

		case INST_MOVEPIXEL64:			// r0 <- r1
			stack[fp+inst.r0+0] = stack[fp+inst.r1+0];
			stack[fp+inst.r0+1] = stack[fp+inst.r1+1];
			break;
		case INST_INVPIXEL64:
			stack[fp+inst.r0+0] = ~stack[fp+inst.r1+0];
			stack[fp+inst.r0+1] = ~stack[fp+inst.r1+1];
			break;
		case INST_ADDPIXEL64:			// r0 <- r1 + r2
			u = stack[fp+inst.r1+1];
			v = stack[fp+inst.r1];
			stack[fp+inst.r0+0] = ((v & 0x7fff7fff) + stack[fp+inst.r2+0]) ^ (v & 0x80008000);
			stack[fp+inst.r0+1] = ((u & 0x7fff7fff) + stack[fp+inst.r2+1]) ^ (u & 0x80008000);
			break;
		case INST_SUBPIXEL64:			// r0 <- r1 - r2
			u = stack[fp+inst.r1+1];
			v = stack[fp+inst.r1];
			stack[fp+inst.r0+0] = ((v | 0x80008000) + stack[fp+inst.r2+0]) ^ (v & 0x80008000);
			stack[fp+inst.r0+1] = ((u | 0x80008000) + stack[fp+inst.r2+1]) ^ (u & 0x80008000);
			break;
		case INST_MULPIXEL64:			// r0 <- r1 * r2
			u = stack[fp+inst.r1];
			v = stack[fp+inst.r2];
			stack[fp+inst.r0+1]	= ((long)(signed short)stack[fp+inst.r1+1] * (long)(signed short)stack[fp+inst.r2+1]);
			stack[fp+inst.r0+0] = (((long)stack[fp+inst.r1+0]>>16) * ((long)stack[fp+inst.r2+0]>>16))
								+ (((long)(signed short)stack[fp+inst.r1+0] * (long)(signed short)stack[fp+inst.r2+0])&0xffff);
			break;
		case INST_ANDPIXEL64:
			stack[fp+inst.r0+0] = stack[fp+inst.r1+0] & stack[fp+inst.r2+0];
			stack[fp+inst.r0+1] = stack[fp+inst.r1+1] & stack[fp+inst.r2+1];
			break;
		case INST_ORPIXEL64:
			stack[fp+inst.r0+0] = stack[fp+inst.r1+0] | stack[fp+inst.r2+0];
			stack[fp+inst.r0+1] = stack[fp+inst.r1+1] | stack[fp+inst.r2+1];
			break;
		case INST_XORPIXEL64:
			stack[fp+inst.r0+0] = stack[fp+inst.r1+0] ^ stack[fp+inst.r2+0];
			stack[fp+inst.r0+1] = stack[fp+inst.r1+1] ^ stack[fp+inst.r2+1];
			break;
		case INST_LSHIFTPIXEL64:
			u = stack[fp + inst.r2];
			((short *)(stack + fp + inst.r0))[0] = ((short *)(stack + fp + inst.r1))[0] << u;
			((short *)(stack + fp + inst.r0))[1] = ((short *)(stack + fp + inst.r1))[1] << u;
			((short *)(stack + fp + inst.r0))[2] = ((short *)(stack + fp + inst.r1))[2] << u;
			break;
		case INST_RSHIFTPIXEL64:
			u = stack[fp + inst.r2];
			((short *)(stack + fp + inst.r0))[0] = ((short *)(stack + fp + inst.r1))[0] >> u;
			((short *)(stack + fp + inst.r0))[1] = ((short *)(stack + fp + inst.r1))[1] >> u;
			((short *)(stack + fp + inst.r0))[2] = ((short *)(stack + fp + inst.r1))[2] >> u;
			break;
		case INST_RLSHIFTPIXEL64:
			u = stack[fp + inst.r2];
			((unsigned short *)(stack + fp + inst.r0))[0] = ((unsigned short *)(stack + fp + inst.r1))[0] >> u;
			((unsigned short *)(stack + fp + inst.r0))[1] = ((unsigned short *)(stack + fp + inst.r1))[1] >> u;
			((unsigned short *)(stack + fp + inst.r0))[2] = ((unsigned short *)(stack + fp + inst.r1))[2] >> u;
			break;
		case INST_GETRPIXEL64:			// r0 <- r1.r
			stack[fp+inst.r0] = ((stack[fp+inst.r1+1] & 0xffff) + 0xffff8000) ^ 0xffff8000;
			break;
		case INST_GETGPIXEL64:			// r0 <- r1.g
			stack[fp+inst.r0] = stack[fp+inst.r1+0] >> 16;
			break;
		case INST_GETBPIXEL64:			// r0 <- r1.b
			stack[fp+inst.r0] = ((stack[fp+inst.r1+0] & 0xffff) + 0xffff8000) ^ 0xffff8000;
			break;
		case INST_SETRPIXEL64:			// r0.r <- r1
			stack[fp+inst.r0+1] = (stack[fp+inst.r0+1] & 0xffff0000) + (stack[fp+inst.r1]&0xffff);
			break;
		case INST_SETGPIXEL64:			// r0.g <- r1
			stack[fp+inst.r0] = (stack[fp+inst.r0] & 0x0000ffff) + (stack[fp+inst.r1]<<16);
			break;
		case INST_SETBPIXEL64:			// r0.b <- r1
			stack[fp+inst.r0] = (stack[fp+inst.r0] & 0xffff0000) + (stack[fp+inst.r1]&0xffff);
			break;
		case INST_PIXEL32TOPIXEL64:		// r0 <- r1
			u = stack[fp+inst.r1];
			stack[fp+inst.r0+0] = ((u&0x0000ff00)<<8) + (u&0x000000ff);
			stack[fp+inst.r0+1] = ((u&0xff000000)>>8) + ((u&0x00ff0000)>>16);
			break;

		// Integer instructions

		case INST_CONSTINT:				// r0 <- #r1
			stack[fp+inst.r0] = inst.r1;
			break;
		case INST_MOVEINT:				// r0 <- r1
			stack[fp+inst.r0] = stack[fp+inst.r1];
			break;
		case INST_NEGINT:				// r0 <- -r1
			stack[fp+inst.r0] = -stack[fp+inst.r0];
			break;
		case INST_INVINT:				// r0 <- ~r1
			stack[fp+inst.r0] = ~stack[fp+inst.r0];
			break;
		case INST_INCINT:				// ++r0;
			++stack[fp+inst.r0];
			break;
		case INST_DECINT:				// --r0;
			--stack[fp+inst.r0];
			break;
		case INST_ADDINT:				// r0 <- r1 + r2
			stack[fp+inst.r0] = stack[fp+inst.r1] + stack[fp+inst.r2];
			break;
		case INST_ADDCONSTINT:			// r0 <- r1 + #r2
			stack[fp+inst.r0] = stack[fp+inst.r1] + inst.r2;
			break;
		case INST_SUBINT:				// r0 <- r1 - r2
			stack[fp+inst.r0] = stack[fp+inst.r1] - stack[fp+inst.r2];
			break;
		case INST_MULINT:				// r0 <- r1 * r2
			stack[fp+inst.r0] = stack[fp+inst.r1] * stack[fp+inst.r2];
			break;
		case INST_DIVINT:				// r0 <- r1 / r2
			stack[fp+inst.r0] = stack[fp+inst.r1] / stack[fp+inst.r2];
			break;
		case INST_ANDINT:
			stack[fp+inst.r0] = stack[fp+inst.r1] & stack[fp+inst.r2];
			break;
		case INST_ORINT:
			stack[fp+inst.r0] = stack[fp+inst.r1] | stack[fp+inst.r2];
			break;
		case INST_XORINT:
			stack[fp+inst.r0] = stack[fp+inst.r1] ^ stack[fp+inst.r2];
			break;
		case INST_LSHIFTINT:
			stack[fp+inst.r0] = stack[fp+inst.r1] << stack[fp+inst.r2];
			break;
		case INST_RSHIFTINT:
			stack[fp+inst.r0] = stack[fp+inst.r1] >> stack[fp+inst.r2];
			break;
		case INST_RLSHIFTINT:
			stack[fp+inst.r0] = (int)((unsigned int)stack[fp+inst.r1] >> stack[fp+inst.r2]);
			break;
		case INST_LANDINT:				// r0 <- r1 && r2
			stack[fp+inst.r0] = stack[fp+inst.r1] && stack[fp+inst.r2];
			break;
		case INST_LORINT:				// r0 <- r1 || r2
			stack[fp+inst.r0] = stack[fp+inst.r1] || stack[fp+inst.r2];
			break;
		case INST_CMPLTINT:				// r0 <- r1 < r2
			stack[fp+inst.r0] = stack[fp+inst.r1] < stack[fp+inst.r2];
			break;
		case INST_CMPLEINT:				// r0 <- r1 <= r2
			stack[fp+inst.r0] = stack[fp+inst.r1] <= stack[fp+inst.r2];
			break;
		case INST_CMPGTINT:				// r0 <- r1 > r2
			stack[fp+inst.r0] = stack[fp+inst.r1] > stack[fp+inst.r2];
			break;
		case INST_CMPGEINT:				// r0 <- r1 >= r2
			stack[fp+inst.r0] = stack[fp+inst.r1] >= stack[fp+inst.r2];
			break;
		case INST_CMPEQINT:				// r0 <- r1 == r2
			stack[fp+inst.r0] = stack[fp+inst.r1] == stack[fp+inst.r2];
			break;
		case INST_CMPNEINT:				// r0 <- r1 != r2
			stack[fp+inst.r0] = stack[fp+inst.r1] != stack[fp+inst.r2];
			break;
		default:
			__assume(false);
		}
	}

finished:
	r1 = return1;
	r2 = return2;
}
