#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <commctrl.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"
/*
*******************************************************************************

	Virtualdub filter to correct colours brightness and contrast
	-----------------------------------------------------------


		Copyright (C) 2002-2004 Alessandro Malanca, 

		http://web.tiscali.it/minomala/
 
	-----------------------------------------------------------

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

*******************************************************************************
*/

int  coloursRunProc ( const FilterActivation *fa, const FilterFunctions *ff);
int  coloursStartProc ( FilterActivation *fa, const FilterFunctions *ff);
int  coloursEndProc ( FilterActivation *fa, const FilterFunctions *ff);
long coloursParamProc ( FilterActivation *fa, const FilterFunctions *ff);
int  coloursConfigProc ( FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void coloursStringProc ( const FilterActivation *fa, const FilterFunctions *ff, char *str);
void coloursScriptConfig ( IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool coloursFssProc ( FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);
int  coloursInitProc ( FilterActivation *fa, const FilterFunctions *ff);


///////////////////////////////////////////////////////////////////////////


typedef struct BrightnessContrast {

	int		brightness;
	int		contrast;

} BrightnessContrast;

typedef struct FilterParameters {

	BrightnessContrast red,global,blue;
	int		contrastBalance;

} FilterParameters;


typedef enum ShowingSelection  {Global, Red, Green, Blue } ShowingSelection;


class FrameHistogram {

#define NUM_SAMPLES 80

public:

void setShowingSelection ( ShowingSelection s ) {
	showingSelection = s;
}

void nextShowingSelection() {
	if (showingSelection == Blue) {
		showingSelection = Global;
	} else {
		showingSelection = (ShowingSelection) ((int)(showingSelection)+1);
	}
}


int getMaxHistoValue() {
	return maxHistoValue;
}


void  paintHisto (HWND hdlg) {

	RECT	rHistoCtrl;

	HDC hdc;
	PAINTSTRUCT ps;

	HPEN whitePen = (HPEN)GetStockObject(WHITE_PEN);
	HPEN blackPen = (HPEN)GetStockObject(BLACK_PEN);

	hdc = BeginPaint(GetDlgItem(hdlg,IDC_HISTO), &ps);

	GetClientRect(GetDlgItem(hdlg, IDC_HISTO), &rHistoCtrl);
	paintBackGr(hdc,rHistoCtrl);

	RECT rHisto;
	int lineSize = calcPaintArea (rHisto, rHistoCtrl );
	FillRect(hdc, &rHisto, (HBRUSH)GetStockObject(LTGRAY_BRUSH));

	HPEN pen;

	switch(showingSelection) {
		case Red:	
			pen = CreatePen( PS_SOLID, 1, RGB(200,0,0));
			break;
		case Blue:	
			pen = CreatePen( PS_SOLID, 1, RGB(0,0,200));
			break;
		case Green:	
			pen = CreatePen( PS_SOLID, 1, RGB(0,240,0));
			break;
		default:
			pen = whitePen;
	}
	
	int xPos = rHisto.left;

	SelectObject( hdc, pen);

	for ( int j=0; j<NUM_SAMPLES; j++ ) {

		if ( j == histoMediaBalance ) {
			SelectObject( hdc, blackPen);
		} else {
			SelectObject( hdc, pen);
		}

		int h = scaleValue(histogram[j],rHisto.bottom-rHisto.top);
		for ( int l=0; l<lineSize; l++) {
			MoveToEx(hdc, xPos+l, rHisto.bottom, NULL);		
			LineTo(hdc,	xPos+l, rHisto.bottom-h);
		}
		xPos += lineSize;
	}

	DeleteObject(pen); 
	EndPaint(hdlg, &ps);
	return;
}


void invalidate(HWND hdlg) {
	InvalidateRgn(hdlg,NULL,FALSE);
}

void calculateHisto( const FilterActivation *fa, int mediaBalance ) {

	int w = fa->dst.w;
	int h = fa->dst.h;

	Pixel32 *dst =  (Pixel32 *)fa->dst.data;

	long summs[256]; 

	for (int j=0; j<256; j++) { summs[j] = 0; }

    int y = h;
	do {int x = w;
		do {
			Pixel32 pixel = *dst++;

			int v;

			switch (showingSelection) {
				case Global:
					v = ((pixel & 0xFF0000)>>16) +
						((pixel & 0x00FF00)>>8) +
						(pixel & 0x0000FF);
					v /= 3;
					break;
				case Red:
					v = (pixel & 0xFF0000)>>16;
					break;
				case Green:
					v = (pixel & 0x00FF00)>>8;
					break;
				case Blue:
					v = (pixel & 0x0000FF);
					break;
			}
			summs[v]++;	
		} while(--x);
		dst = (Pixel32 *)((char *)dst + fa->dst.modulo);
    } while(--y);

	double max = 0; int next = 0;
	int step = 256/NUM_SAMPLES;
	for( j=0; j<NUM_SAMPLES; j++) {
		histogram[j] = 0;
		for ( int k=0; k < step ; k++ ) {
			if ( next == mediaBalance ) {
				histoMediaBalance = j;
			}
			histogram[j] += summs[next++];
			if (histogram[j] > max) max = histogram[j];
		}
	}

	for( j=0; j<NUM_SAMPLES; j++) {
		histogram[j] = histogram[j] / max; 
	}

	maxHistoValue = (long)max;
	// in 0,0 .. 1,0
}

private:

	double		histogram[NUM_SAMPLES];
	ShowingSelection	showingSelection;
	long		maxHistoValue;
	long		histoMediaBalance;

void paintBackGr( HDC hdc, RECT r ) {

	FillRect(hdc, &r, (HBRUSH)GetStockObject(LTGRAY_BRUSH));

	HPEN blackPen = (HPEN)GetStockObject(BLACK_PEN);
	SelectObject( hdc, blackPen);

	MoveToEx(hdc, r.left, r.top, NULL);
	LineTo(hdc,	r.right-1,r.top);
	LineTo(hdc,	r.right-1,r.bottom-1);
	LineTo(hdc,	r.left,r.bottom-1);
	LineTo(hdc,	r.left,r.top);

}

int calcPaintArea ( RECT & res, RECT rCtrl ) {

	int step = (rCtrl.right-rCtrl.left) / NUM_SAMPLES;
	int left = (rCtrl.right - rCtrl.left - (step * NUM_SAMPLES)) / 2;
	res.left = rCtrl.left+left;
	res.right = rCtrl.left+left+step*NUM_SAMPLES;
	res.bottom = rCtrl.bottom-4;
	res.top = rCtrl.top+4;
	return step;

}

int scaleValue(double histo, int max) {

	// histo in 0..1

	if ( histo <= 0 ) return 0;
	if ( histo >= 1 ) return max;
	return (int) ( histo * (double) max );

}

}; // FrameHistogram


typedef struct PixelAdjust {

	Pixel32 red[256];
	Pixel32 green[256];
	Pixel32 blue[256];

} PixelAdjust;


typedef struct MyFilterData {
	
	bool				disableFilter;
	FilterParameters	par,save;
	IFilterPreview		*ifp;
	FrameHistogram      frameHistogram;
	bool				displayingPanel;
	PixelAdjust			*pixelAdjust; 

} MyFilterData;


ScriptFunctionDef colours_func_defs[]={
    { (ScriptFunctionPtr)coloursScriptConfig, "Config", "0iiiiii" }, 
    { NULL },
};

CScriptObject colours_obj = {
    NULL, colours_func_defs
};

struct FilterDefinition filterDef_colours = {

    NULL, NULL, NULL,			// next, prev, module
    "colours brightness contrast [1.0]",	// name
    "rgb colours brightness contrast correction",	// desc
    "Alessandro Malanca",		// maker
    NULL,						// private_data
    sizeof(MyFilterData),		// inst_data_size
    coloursInitProc,			// initProc
    NULL,						// deinitProc
    coloursRunProc,				// runProc
    NULL,						// ParamProc
    coloursConfigProc,			// ConfigProc
    coloursStringProc,			// stringProc
    coloursStartProc,			// startProc
    coloursEndProc,				// endProc
    &colours_obj,				// script_obj
    coloursFssProc,				// fssProc

};

///////////////////////////////////////////////////////////////////////////

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_colours;

///////////////////////////////////////////////////////////////////////////

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_colours = ff->addFilter(fm, &filterDef_colours, sizeof(FilterDefinition))))
        return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_colours);
}

///////////////////////////////////////////////////////////////////////////


int coloursInitProc(FilterActivation *fa, const FilterFunctions *ff) {

	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	mfd->displayingPanel = false;
	mfd->par.contrastBalance = 120;
	mfd->frameHistogram.setShowingSelection(Red);
	return 0;

}

#define MAX_CONTRAST	100
#define MAX_BRIGHTNESS	100

PixelAdjust * prepareToAdjust(FilterParameters par) {

	long r,g,b;
	
	PixelAdjust * res;

	res = new PixelAdjust();

	for (int i=0; i<256; i++) {

		r = g = b = i;

		// --------------

		// adjust colours contrast

		long red_contr = 256 + par.global.contrast + par.red.contrast;
		long gre_contr = 256 + par.global.contrast;
		long blu_contr = 256 + par.global.contrast + par.blue.contrast;

		r = ( ( (r-par.contrastBalance) * red_contr ) / 256) + par.contrastBalance;
		g = ( ( (g-par.contrastBalance) * gre_contr ) / 256) + par.contrastBalance;
		b = ( ( (b-par.contrastBalance) * blu_contr ) / 256) + par.contrastBalance;

		int red_bri = par.global.brightness + par.red.brightness;
		int gre_bri = par.global.brightness;
		int blu_bri = par.global.brightness + par.blue.brightness;

		red_bri /= 2;
		gre_bri /= 2;
		blu_bri /= 2;

		// additive part
		r += red_bri;
		g += gre_bri;
		b += blu_bri;


		// multiply part
		if ( red_bri > 0 ) red_bri *= 2;
		if ( gre_bri > 0 ) gre_bri *= 2;
		if ( blu_bri > 0 ) blu_bri *= 2;

		red_bri += 256;
		gre_bri += 256;
		blu_bri += 256; 

		r = ( r * red_bri ) / 256;
		g = ( g * gre_bri ) / 256;
		b = ( b * blu_bri ) / 256;

		if (r < 0) r=0; else if (r > 255) r=255;
		if (g < 0) g=0; else if (g > 255) g=255;
		if (b < 0) b=0; else if (b > 255) b=255;

		res->red[i]	  = r<<16;
		res->green[i] = g<<8;
		res->blue[i]  = b;

	}

	return res;

}


int coloursStartProc(FilterActivation *fa, const FilterFunctions *ff) {

    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->pixelAdjust = prepareToAdjust(mfd->par);

    return 0;

}


int coloursRunProc(const FilterActivation *fa, const FilterFunctions *ff) {

    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    PixDim w, h;
    Pixel32 *src, *dst;

	if (mfd->disableFilter) {
		fa->dst.data = fa->src.data;
		mfd->frameHistogram.calculateHisto(fa, mfd->par.contrastBalance);
		return 0;
	}

    src = (Pixel32 *)fa->src.data;
    dst = (Pixel32 *)fa->dst.data;

	Pixel32 * red_pix = mfd->pixelAdjust->red;
	Pixel32 * gre_pix = mfd->pixelAdjust->green;
	Pixel32 * blu_pix = mfd->pixelAdjust->blue;

    h = fa->src.h;
    do { 
		w = fa->src.w;
		do {

			// adjust colors brightness and contrast 

			*dst++ =  red_pix [ (*src & 0xFF0000) >> 16 ] 
					| gre_pix [ (*src & 0x00FF00) >> 8 ] 
					| blu_pix [ (*src & 0x0000FF) ];

			src++;

		} while(--w);

	    src = (Pixel32 *)((char *)src + fa->src.modulo);
		dst = (Pixel32 *)((char *)dst + fa->dst.modulo);

    } while(--h);

	if (mfd->displayingPanel) {
		mfd->frameHistogram.calculateHisto(fa, mfd->par.contrastBalance);
	}

    return 0;
}


int coloursEndProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	if ( mfd->pixelAdjust != NULL ) {
		delete mfd->pixelAdjust;
		mfd->pixelAdjust=NULL;
	}

    return 0;
}


void setDialogFeedbackValues (HWND hdlg, MyFilterData *mfd) {

	SetDlgItemInt(hdlg, IDC_RED_BRIGHT_VALUE, mfd->par.red.brightness, TRUE);
	SetDlgItemInt(hdlg, IDC_RED_CONTR_VALUE, mfd->par.red.contrast, TRUE);
	SetDlgItemInt(hdlg, IDC_GLOB_BRIGHT_VALUE, mfd->par.global.brightness, TRUE);
	SetDlgItemInt(hdlg, IDC_GLOB_CONTR_VALUE, mfd->par.global.contrast, TRUE);
	SetDlgItemInt(hdlg, IDC_BLUE_BRIGHT_VALUE, mfd->par.blue.brightness, TRUE);
	SetDlgItemInt(hdlg, IDC_BLUE_CONTR_VALUE, mfd->par.blue.contrast, TRUE);

}


void redoSystem( MyFilterData * mfd, HWND hdlg) {		

	mfd->ifp->UndoSystem();
	mfd->ifp->RedoSystem();
	setDialogFeedbackValues(hdlg,mfd);
	mfd->frameHistogram.invalidate(hdlg);
}


void redoFrame ( MyFilterData * mfd, HWND hdlg ) {

	mfd->ifp->RedoFrame();
	setDialogFeedbackValues(hdlg,mfd);
	mfd->frameHistogram.invalidate(hdlg);

}

void getHorSliderValues(MyFilterData *mfd,HWND hdlg) {

	// GLOBAL
	int global_bright = SendMessage(GetDlgItem(hdlg, IDC_GLOB_BRIGHT_SLIDER), TBM_GETPOS, 0, 0);
	if ( global_bright != mfd->par.global.brightness ) {
		mfd->par.global.brightness = global_bright;
		mfd->frameHistogram.setShowingSelection(Global);
		redoSystem( mfd, hdlg );
		return ;
	}
	int global_contr = SendMessage(GetDlgItem(hdlg, IDC_GLOB_CONTR_SLIDER), TBM_GETPOS, 0, 0);
	if ( global_contr != mfd->par.global.contrast ) {
		mfd->par.global.contrast = global_contr;
		mfd->frameHistogram.setShowingSelection(Global);
		redoSystem( mfd, hdlg );
		return ;
	}
	
	// RED
	int red_bright = SendMessage(GetDlgItem(hdlg, IDC_RED_BRIGHT_SLIDER), TBM_GETPOS, 0, 0);
	if ( red_bright != mfd->par.red.brightness ) {
		mfd->par.red.brightness = red_bright;
		mfd->frameHistogram.setShowingSelection(Red);
		redoSystem( mfd, hdlg );
		return ;
	}
	int red_contr = SendMessage(GetDlgItem(hdlg, IDC_RED_CONTR_SLIDER), TBM_GETPOS, 0, 0);
	if ( red_contr != mfd->par.red.contrast ) {
		mfd->par.red.contrast = red_contr;
		mfd->frameHistogram.setShowingSelection(Red);
		redoSystem( mfd, hdlg );
		return ;
	}

	// BLUE
	int blue_bright = SendMessage(GetDlgItem(hdlg, IDC_BLUE_BRIGHT_SLIDER), TBM_GETPOS, 0, 0);
	if ( blue_bright != mfd->par.blue.brightness ) {
		mfd->par.blue.brightness = blue_bright;
		mfd->frameHistogram.setShowingSelection(Blue);
		redoSystem( mfd, hdlg );
		return ;
	}
	int blue_contr = SendMessage(GetDlgItem(hdlg, IDC_BLUE_CONTR_SLIDER), TBM_GETPOS, 0, 0);
	if ( blue_contr != mfd->par.blue.contrast ) {
		mfd->par.blue.contrast = blue_contr;
		mfd->frameHistogram.setShowingSelection(Blue);
		redoSystem( mfd, hdlg );
		return ;
	}

}


static void levelsButtonCallback(bool fNewState, void *pvData)
{
	HWND hdlg = (HWND)pvData;
	EnableWindow(GetDlgItem(hdlg, IDSAMPLE), fNewState);
	EnableWindow(GetDlgItem(hdlg, IDC_DISABLE_CHECK), fNewState);
	EnableWindow(GetDlgItem(hdlg, IDC_HISTO_SELECTION), fNewState);

	HWND wHist = GetDlgItem(hdlg, IDC_HISTO);
	if (fNewState == TRUE) {
		ShowWindow(wHist, SW_SHOWNORMAL);
	} else {
		ShowWindow(wHist, SW_HIDE);
	}

}


void setPanelDefaults(HWND hdlg, MyFilterData *mfd) {

	HWND hWnd;

	// GLOBAL
	hWnd = GetDlgItem(hdlg,IDC_GLOB_CONTR_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-MAX_CONTRAST, MAX_CONTRAST));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.global.contrast );

	hWnd = GetDlgItem(hdlg,IDC_GLOB_BRIGHT_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-MAX_BRIGHTNESS, MAX_BRIGHTNESS));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.global.brightness);

	// RED
	hWnd = GetDlgItem(hdlg,IDC_RED_CONTR_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-MAX_CONTRAST, MAX_CONTRAST));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.red.contrast );

	hWnd = GetDlgItem(hdlg,IDC_RED_BRIGHT_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-MAX_BRIGHTNESS, MAX_BRIGHTNESS));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.red.brightness);

	// BLUE
	hWnd = GetDlgItem(hdlg,IDC_BLUE_CONTR_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-MAX_CONTRAST, MAX_CONTRAST));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.blue.contrast );

	hWnd = GetDlgItem(hdlg,IDC_BLUE_BRIGHT_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-MAX_BRIGHTNESS, MAX_BRIGHTNESS));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.blue.brightness);

}



BOOL CALLBACK coloursConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {

    MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);

    switch(msg) {

        case WM_INITDIALOG:
            SetWindowLong(hdlg, DWL_USER, lParam);
            mfd = (MyFilterData *)lParam;

			mfd->displayingPanel = true;
			EnableWindow(GetDlgItem(hdlg, IDSAMPLE), false);

			levelsButtonCallback(false, (void *)hdlg);
			mfd->ifp->SetButtonCallback(levelsButtonCallback, (void *)hdlg);
			SetDlgItemInt(hdlg, IDC_CONTRAST_BALANCE, mfd->par.contrastBalance, TRUE);

			// save for cancel ...
			mfd->save = mfd->par;
			HWND hWnd;

			hWnd = GetDlgItem(hdlg,IDPREVIEW);
			mfd->ifp->InitButton(hWnd);
			setPanelDefaults(hdlg,mfd);

			// update vales in boxes
			setDialogFeedbackValues(hdlg,mfd);

			return TRUE;


        case WM_COMMAND:

            switch(LOWORD(wParam)) {

            case IDOK:
				mfd->disableFilter = false;
				mfd->displayingPanel = false;
                EndDialog(hdlg, 0);
                return TRUE;

			case IDHELP:	{
				char prog[256];
				char path[256];
				LPTSTR ptr;

				GetModuleFileName(NULL, prog, 255);
				GetFullPathName(prog, 255, path, &ptr);
				*ptr = 0;
				strcat(path, "plugins\\ColoursBriCont.htm");
				ShellExecute(hdlg, "open", path, NULL, NULL, SW_SHOWNORMAL);
				return TRUE;
				}

            case IDCANCEL: 
				mfd->par = mfd->save;
				mfd->disableFilter = false;
				mfd->displayingPanel = false;
                EndDialog(hdlg, 1);
                return FALSE;

			case IDPREVIEW: 
				mfd->ifp->RedoSystem();
				mfd->ifp->Toggle(GetDlgItem(hdlg,IDPREVIEW));
				break;
			
			case IDSAMPLE:
				mfd->frameHistogram.invalidate(hdlg);
				break;

			case IDC_BRIGHT_GLOB_DEFAULT :
				mfd->par.global.brightness =0;
				SendMessage(GetDlgItem(hdlg,IDC_GLOB_BRIGHT_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, 0);
				redoSystem( mfd, hdlg );
				break;

			case IDC_BRIGHT_REDBLUE_DEFAULT :
				mfd->par.red.brightness = 
				mfd->par.blue.brightness = 0;
				SendMessage(GetDlgItem(hdlg,IDC_RED_BRIGHT_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, 0);
				SendMessage(GetDlgItem(hdlg,IDC_BLUE_BRIGHT_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, 0);
				redoSystem( mfd, hdlg );
				break;


			case IDC_CONTR_GLOB_DEFAULT :
				mfd->par.global.contrast = 0;
				SendMessage(GetDlgItem(hdlg,IDC_GLOB_CONTR_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, 0);
				redoSystem( mfd, hdlg );
				mfd->par.contrastBalance = 120;
				SetDlgItemInt(hdlg, IDC_CONTRAST_BALANCE, mfd->par.contrastBalance, FALSE);
				break;

			case IDC_CONTR_REDBLUE_DEFAULT :
				mfd->par.red.contrast = 
				mfd->par.blue.contrast = 0;
				SendMessage(GetDlgItem(hdlg,IDC_RED_CONTR_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, 0);
				SendMessage(GetDlgItem(hdlg,IDC_BLUE_CONTR_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, 0);
				redoSystem( mfd, hdlg );
				break;

			case IDC_DISABLE_CHECK:
				mfd->disableFilter = ! mfd->disableFilter;
				mfd->ifp->RedoFrame();
				mfd->frameHistogram.invalidate(hdlg);
				break;

			case IDC_HISTO_SELECTION:
				mfd->frameHistogram.nextShowingSelection();
				mfd->ifp->RedoFrame();
				mfd->frameHistogram.invalidate(hdlg);
				break;


			case IDC_CONTRAST_BALANCE: {
				int balance = GetDlgItemInt(hdlg, IDC_CONTRAST_BALANCE, NULL, FALSE);
				if (balance != mfd->par.contrastBalance) {
					if ( balance > 255 )  {
						balance=255;
						SetDlgItemInt(hdlg, IDC_CONTRAST_BALANCE, balance, FALSE);
					}
					mfd->par.contrastBalance=balance;
					redoSystem( mfd, hdlg );

				}
				} break;

		}	// case WM_COMMAND:

		case WM_HSCROLL: 
			if (LOWORD(wParam) == TB_ENDTRACK)  
				getHorSliderValues(mfd,hdlg);
			break;

		case WM_PAINT:
			SetDlgItemInt(hdlg, IDC_MAX_VALUE, mfd->frameHistogram.getMaxHistoValue(), TRUE);
			mfd->frameHistogram.paintHisto(hdlg);			
			break;
    }

    return FALSE;
}

int coloursConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {

	MyFilterData * mfd = (MyFilterData *)fa->filter_data;
	mfd->ifp = fa->ifp;
	mfd->displayingPanel = false;
    return DialogBoxParam(fa->filter->module->hInstModule,
            MAKEINTRESOURCE(IDD_FILTER_COLOURS), hwnd,
            coloursConfigDlgProc, (LPARAM)fa->filter_data
	);
}


void coloursStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {

    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	sprintf(str,"(global b=%d c=%d, red b=%d c=%d, blue b=%d c=%d )",
		mfd->par.global.brightness, mfd->par.global.contrast,		
		mfd->par.red.brightness, mfd->par.red.contrast,
		mfd->par.blue.brightness, mfd->par.blue.contrast
	);

}

void coloursScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {

    FilterActivation *fa = (FilterActivation *)lpVoid;
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	int ctr=0;
	mfd->par.global.brightness	= argv[ctr++].asInt();
	mfd->par.global.contrast	= argv[ctr++].asInt();
	mfd->par.red.brightness		= argv[ctr++].asInt();
	mfd->par.red.contrast		= argv[ctr++].asInt();
	mfd->par.blue.brightness	= argv[ctr++].asInt();
	mfd->par.blue.contrast		= argv[ctr++].asInt();

}


bool coloursFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    _snprintf(buf, buflen, "Config(%d, %d, %d, %d, %d, %d)",
		mfd->par.global.brightness,	mfd->par.global.contrast,
		mfd->par.red.brightness,	mfd->par.red.contrast,
		mfd->par.blue.brightness,	mfd->par.blue.contrast
	);

    return true;
}
