//	filtfactory - poor man's filter compiler for VirtualDub
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <crtdbg.h>
#include <windows.h>
#include <math.h>
#include <string.h>
#include <stdio.h>

#include "resource.h"
#include "filter.h"
#include "parser.h"
#include "codegen.h"
#include "interp.h"

int filtfactory_init(FilterActivation *fa, const FilterFunctions *ff);
void filtfactory_deinit(FilterActivation *fa, const FilterFunctions *ff);
int filtfactory_run(const FilterActivation *fa, const FilterFunctions *ff);
long filtfactory_param(FilterActivation *fa, const FilterFunctions *ff);
int filtfactory_config(FilterActivation *fa, const FilterFunctions *ff, HWND hWnd);
void filtfactory_string(const FilterActivation *fa, const FilterFunctions *ff, char *buf);
int filtfactory_start(FilterActivation *fa, const FilterFunctions *ff);
int filtfactory_end(FilterActivation *fa, const FilterFunctions *ff);

static FilterDefinition *fd_filtfactory;

class filtfactoryFilterData {
public:
	Instruction *code;
	char *text;
};


static struct FilterDefinition filterDef_filtfactory={
	0,0,NULL,
	"filter factory",
	"Lets you create your own filters in almost-wannabe-C.",
	"Avery Lee",
	NULL,sizeof(filtfactoryFilterData),
	filtfactory_init,
	filtfactory_deinit,
	filtfactory_run,
	filtfactory_param,
	filtfactory_config,
//	filtfactory_string,
	NULL,
	filtfactory_start,
	filtfactory_end,
};


extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
	fd_filtfactory = ff->addFilter(fm, &filterDef_filtfactory, sizeof(FilterDefinition));

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

	return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
	ff->removeFilter(fd_filtfactory);
}

////////////////////////////////////////////////////////////

int filtfactory_init(FilterActivation *fa, const FilterFunctions *ff) {
	filtfactoryFilterData *sfd = (filtfactoryFilterData *)fa->filter_data;

	sfd->text = strdup(
					"void run(bitmap dst, bitmap src) {\r\n"
					"\tint x, y;\r\n"
					"\r\n"
					"\tscan(dst, src) {\r\n"
					"\t\tout = ~in;\r\n"
					"\t}\r\n"
					"}\r\n"
				);

	return 0;
}
void filtfactory_deinit(FilterActivation *fa, const FilterFunctions *ff) {
	filtfactoryFilterData *sfd = (filtfactoryFilterData *)fa->filter_data;

	free(sfd->code);
	free(sfd->text);
}

long filtfactory_param(FilterActivation *fa, const FilterFunctions *ff) {
	return FILTERPARAM_SWAP_BUFFERS;
}

void filtfactory_string(const FilterActivation *fa, const FilterFunctions *ff, char *buf) {
}

/////////////////////////////////////////////////////////////

LRESULT CALLBACK EditWithTabsProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	if (msg == WM_GETDLGCODE) {
		LPMSG pMsg = (LPMSG)lParam;
		LRESULT lres;

//		if (pMsg && ((pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_TAB) || (pMsg->message == WM_CHAR && pMsg->wParam == '\t')))
			return DLGC_WANTTAB | DLGC_HASSETSEL | DLGC_WANTALLKEYS | DLGC_WANTARROWS | DLGC_WANTCHARS;

//		return DLGC_WANTTAB | CallWindowProc((WNDPROC)GetWindowLong(hwnd, GWL_USERDATA), hwnd, msg, wParam, lParam);
	}
	return CallWindowProc((WNDPROC)GetWindowLong(hwnd, GWL_USERDATA), hwnd, msg, wParam, lParam);
}

BOOL APIENTRY filtfactoryConfigDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	filtfactoryFilterData *sfd = (filtfactoryFilterData *)GetWindowLong(hDlg, DWL_USER);

	switch (message)
	{
		case WM_INITDIALOG:
			{
				HWND hwndItem;

				SetWindowLong(hDlg, DWL_USER, lParam);
				sfd = (filtfactoryFilterData *)lParam;

				hwndItem = GetDlgItem(hDlg, IDC_CODE);
				SetWindowLong(hwndItem, GWL_USERDATA, GetWindowLong(hwndItem, GWL_WNDPROC));
				SetWindowLong(hwndItem, GWL_WNDPROC, (LPARAM)EditWithTabsProc);
				SetFocus(GetDlgItem(hDlg, IDC_CODE));

				if (sfd->text)
					SetWindowText(hwndItem, sfd->text);
			}
			return FALSE;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
			case IDC_COMPILE:
				{
					char *buf;
					int len = SendDlgItemMessage(hDlg, IDC_CODE, WM_GETTEXTLENGTH, 0, 0);

					SetFocus(GetDlgItem(hDlg, IDC_CODE));

					if (buf = (char *)malloc(len+1)) {
						if (SendDlgItemMessage(hDlg, IDC_CODE, WM_GETTEXT, len+1, (LPARAM)buf)) {
							try {
								ParseNodeAllocator pn;
								Parser p(&pn);

								TreeTranslationNode *tree = p.Parse(buf);

								CodeGenerator c(&pn, tree);

								if (sfd->code)
									free(sfd->code);

								sfd->code = c.Compile();

								SetDlgItemText(hDlg, IDC_STATIC_MESSAGE, "Compile successful.");
							} catch(ParseError err) {
								int offs = err.getOffset();

								SetDlgItemText(hDlg, IDC_STATIC_MESSAGE, err.getError());

								if (offs >= 0)
									SendDlgItemMessage(hDlg, IDC_CODE, EM_SETSEL, offs, offs);
							}
						}
					}
				}
				return TRUE;
			case IDOK:
				{
					char *buf;
					int len = SendDlgItemMessage(hDlg, IDC_CODE, WM_GETTEXTLENGTH, 0, 0);

					SetFocus(GetDlgItem(hDlg, IDC_CODE));

					if (buf = (char *)malloc(len+1)) {
						if (SendDlgItemMessage(hDlg, IDC_CODE, WM_GETTEXT, len+1, (LPARAM)buf)) {
							free(sfd->text);
							sfd->text = buf;
						}
					}
				}
				EndDialog(hDlg, TRUE);
				return TRUE;

			case IDCANCEL:
				EndDialog(hDlg, FALSE);
				return TRUE;
			}
			break;
	}
	return FALSE;
}

int filtfactory_config(FilterActivation *fa, const FilterFunctions *ff, HWND hWnd) {
	DialogBoxParam(fa->filter->module->hInstModule, MAKEINTRESOURCE(IDD_FILTER_FILTFACTORY), hWnd, (DLGPROC)filtfactoryConfigDlgProc, (LPARAM)fa->filter_data);

	return 0;
}

/////////////////////////////////////////////////////////////

int filtfactory_start(FilterActivation *fa, const FilterFunctions *ff) {
	filtfactoryFilterData *sfd = (filtfactoryFilterData *)fa->filter_data;

	return 0;
}

int filtfactory_run(const FilterActivation *fa, const FilterFunctions *ff) {
	filtfactoryFilterData *sfd = (filtfactoryFilterData *)fa->filter_data;

	long stack[500];

	if (sfd->code) {
		long r1, r2;

		stack[0] = (long)(VBitmap *)&fa->src;
		stack[1] = (long)(VBitmap *)&fa->dst;
		stack[2] = -1;
		stack[3] = 0;

		InterpretCode(sfd->code, 0, stack, 4, r1, r2);
	}

	return 0;
}

int filtfactory_end(FilterActivation *fa, const FilterFunctions *ff) {
	filtfactoryFilterData *sfd = (filtfactoryFilterData *)fa->filter_data;

	return 0;
}
